"""
This test logging module configures test case logging to print
debug messages to stdout.
"""
# Absolute import (the default in a future Python release) resolves
# the logging import as the Python standard logging module rather
# than this module of the same name.
from __future__ import absolute_import
import logging
from qiutil.logging import (configure, logger)

configure('test', 'qipipe', 'qixnat', 'qidicom', 'qiutil', level='DEBUG')

# Suppress warnings, esp. the dependent libraries comparison to None.
logging.captureWarnings(True)



import os
import glob
import shutil
from collections import defaultdict
from nose.tools import (assert_equal, assert_not_equal)
from qiutil.collections import nested_defaultdict
from ... import PROJECT
from ...helpers.logging import logger


class VolumeTestBase(object):
    """
    Base class for testing image volume workflows. The test fixture is
    a directory in the standard staging input subject/session/images
    hierarchy found in the ``test/fixtures/staging`` directory.
    """

    def __init__(self, logger, fixtures, results, use_mask=False):
        """
        :param logger: this test case's logger
        :param fixtures: the fixtures parent directory
        :param results: the results parent directory
        :param use_mask: flag indicating whether the inputs include
            a mask with the input images
        """
        self._logger = logger
        self._fixtures = fixtures
        self._results = results
        self.base_dir = os.path.join(self._results, 'work')
        self._use_mask = use_mask

    def setUp(self):
        shutil.rmtree(self._results, True)

    def tearDown(self):
        shutil.rmtree(self._results, True)

    def stage(self, collection):
        """
        Acquires the test fixture inputs to run the workflow. This
        generator method yields the workflow input arguments,
        consisting of the subject, session and scan names followed
        by the image file names.

        :param collection: the collection name
        :param opts: additional workflow options
        :yield: the workflow inputs list
        """
        # The fixture is the collection subdirectory.
        fixture = os.path.join(self._fixtures, collection.lower())
        # The staging inputs.
        inputs = self._group_files(fixture)
        # The test subjects.
        subjects = inputs.keys()
        # Run the workflow in the results work subdirectory.
        cwd = os.getcwd()
        work_dir = os.path.join(self._results, 'work')
        os.makedirs(work_dir)
        os.chdir(work_dir)
        # Iterate over the inputs.
        try:
            return self._stage_inputs(inputs)
        finally:
            os.chdir(cwd)

    def _stage_inputs(self, inputs):
        """
        :param inputs: the :meth:`_group_files` result dictionary
        :yield: the workflow inputs list
        """
        # Iterate over the sessions within subjects.
        for sbj, sess_dict in inputs.iteritems():
            for sess, scan_dict in sess_dict.iteritems():
                for scan, opts in scan_dict.iteritems():
                    # The input images.
                    images = opts.pop('images')
                    args = [PROJECT, sbj, sess, scan] + images
                    yield args

    def _group_files(self, fixture):
        """
        Groups the files in the given test fixture directory. The fixture is a
        parent directory which contains a subject/session/scans images hierarchy.
        
        :param fixture: the input data parent directory
        :return: the input *{subject: {session: {scan: [images]}}}* or
            *{subject: {session: {scan: ([images], mask: file)}}}*
            dictionary
        """
        # The inputs dictionary return value described above.
        inputs = nested_defaultdict(dict, 2)
        # Group the files in each subject subdirectory.
        for sbj_dir in glob.glob(fixture + '/*'):
            _, sbj = os.path.split(sbj_dir)
            self._logger.debug("Discovered test input subject %s in %s" %
                               (sbj, sbj_dir))
            for sess_dir in glob.glob(sbj_dir + '/Session*'):
                _, sess = os.path.split(sess_dir)
                for scan_dir in glob.glob(sess_dir + '/scans/*'):
                    _, scan_s = os.path.split(scan_dir)
                    scan = int(scan_s)
                    # The input NIFTI volume images.
                    images = glob.glob(scan_dir + '/resources/NIFTI/*')
                    assert_not_equal(len(images), 0,
                                     "No images found for %s %s %d test input in %s" %
                                     (sbj, sess, scan, scan_dir))
                    self._logger.debug("Discovered %d %s %s %d test input images in %s" %
                                       (len(images), sbj, sess, scan, scan_dir))
                    inputs[sbj][sess][scan]['images'] = images
                    # The input mask.
                    if self._use_mask:
                        masks = glob.glob(sess_dir + '/resources/*mask.*')
                        assert_not_equal(len(masks), 0, "Mask not found in %s" % sess_dir)
                        assert_equal(len(masks), 1, "Too many masks found in %s" % sess_dir)
                        mask = masks[0]
                        self._logger.debug("Discovered %d %s %s test input mask %s" %
                                           (len(images), sbj, sess, mask))
                        inputs[sbj][sess][scan]['mask'] = mask

        return inputs



"""
This module updates the qiprofile database imaging information
from a XNAT experiment.
"""
import qixnat
from ..helpers.constants import (SUBJECT_FMT, SESSION_FMT)
from . import modeling


class ImagingError(Exception):
    pass


def sync(subject, session):
    """
    Updates the imaging content for the given qiprofile subject database
    object from the given XNAT experiment.

    :param subject: the target qiprofile subject object to update
    :param session: the session number
    """
    # Sync the content in a XNAT connection context.
    with qixnat.connect() as xnat:
        # The XNAT subject name.
        sbj_nm = SUBJECT_FMT % (collection, subject)
        # The XNAT experiment name.
        exp_nm = SESSION_FMT % session
        # Find the XNAT experiment
        exp = xnat.find_one(project=project, subject=sbj_nm,
                                  experiment=exp_nm)
        if not exp.exists():
            raise ImagingError(
                "%s %s Subject %d Session %d XNAT experiment not found" %
                (subject.project, subject.collection, subject.number,
                 session.number)
            )
        # The XNAT experiment must have a date.
        if not exp.date:
            raise ImagingError(
                "%s %s Subject %d Session %d XNAT experiment is missing"
                " a date" % (subject.project, subject.collection,
                             subject.number, session.number)
            )
        # If there is a qiprofile session with the same date,
        # then complain.
        is_dup_session = lambda sess: sess.date == exp.date
        if any(is_dup_session, sbj.sessions):
            raise ImagingError(
                "qiprofile %s %s Subject %d session with date %s already"
                " exists" % (subject.project, subject.collection,
                             subject.number, exp.date)
            )
        # Make the qiprofile session database object.
        sess = Session(date=exp.date)
        # Add the session to the subject encounters in date order.
        sbj.add_encounter(sess)
        # Update the qiprofile session object.
        _update(sess, exp)
        # Save the session detail.
        sess.detail.save()
        # Save the subject.
        sbj.save()

def _update(session, experiment):
    """
    Updates the qiprofile session from the XNAT experiment.
    
    :param session: the qiprofile session object
    :param experiment: the XNAT experiment object
    """
    # The modeling resources begin with 'pk+'.
    for xrsc in experiment.resources():
        if xrsc.label.startswith('pk_'):
            modeling.update(session, xrsc)
    
    # Create the session detail database subject to hold the scans.
    session.detail = database.get_or_create(SubjectDetail)
    # The scans are embedded in the SessionDetail document.
    for xscan in experiment.scans():
        scan.update(session.detail, xscan)



import os
import re
import glob
import shutil
from nose.tools import assert_is_not_none
import nipype.pipeline.engine as pe
from nipype.interfaces.dcmstack import MergeNifti
import qixnat
# Modeling requires the proprietary OHSU fastfit module.
# If fastfit is not found, then the tests are skipped.
try:
    import fastfit
except ImportError:
    fastfit = None
from qipipe.pipeline import modeling
from qipipe.pipeline import qipipeline
from ... import (ROOT, PROJECT)
from ...helpers.logging import logger
from ...unit.pipeline.volume_test_base import VolumeTestBase

MODELING_CONF = os.path.join(ROOT, 'conf', 'modeling.cfg')
"""The test registration configuration."""

FIXTURES = os.path.join(ROOT, 'fixtures', 'staged')

RESULTS = os.path.join(ROOT, 'results', 'pipeline', 'modeling')
"""The test results directory."""


class TestModelingWorkflow(VolumeTestBase):
    """
    Modeling workflow unit tests.
    This test exercises the modeling workflow on the QIN Breast and Sarcoma
    study visits in the ``test/fixtures/pipeline/modeling`` test fixture
    directory.
    
    Note:: a precondition for running this test is that the
        ``test/fixtures/pipeline/modeling`` directory contains the series
        stack test data in collection/subject/session format, e.g.::
    
            breast
                Breast003
                    Session01
                        volume009.nii.gz
                        volume023.nii.gz
                         ...
            sarcoma
                Sarcoma001
                    Session01
                        volume011.nii.gz
                        volume013.nii.gz
                         ...
    
        The fixture is not included in the Git source repository due to
        storage constraints.
    
    Note:: this test takes several hours to run on the AIRC cluster.
    """

    def __init__(self):
        super(TestModelingWorkflow, self).__init__(
            logger(__name__), FIXTURES, RESULTS)

    def test_breast(self):
        if fastfit:
            for args in self.stage('Breast'):
                self._test_workflow(*args)
        else:
            logger(__name__).debug('Skipping modeling test since fastfit'
                                   ' is unavailable.')

    def test_sarcoma(self):
        if fastfit:
            for args in self.stage('Sarcoma'):
                self._test_workflow(*args)

    def _test_workflow(self, project, subject, session, scan, *images):
        """
        Executes :meth:`qipipe.pipeline.modeling.run` on the input scans.
        
        :param project: the input project name
        :param subject: the input subject name
        :param session: the input session name
        :param scan: the input scan number
        :param images: the input 3D NiFTI images to model
        """
        # Make the 4D time series from the test fixture inputs.
        # TODO - newer nipype has out_path MergeNifti input field. Set
        # that field to out_path=RESULTS below. Work-around is to move
        # the file to RESULTS below.
        merge = MergeNifti(in_files=list(images),
                           out_format=qipipeline.SCAN_TS_RSC)
        time_series = merge.run().outputs.out_file
        # Work-around for nipype bug described above.
        _, ts_fname = os.path.split(time_series)
        ts_dest = os.path.join(RESULTS, ts_fname)
        import shutil
        shutil.move(time_series, ts_dest)
        time_series = ts_dest
        # End of work-around.
        logger(__name__).debug("Testing the modeling workflow on the %s %s"
                               " time series %s..." %
                               (subject, session, time_series))
        with qixnat.connect() as xnat:
            xnat.delete(project, subject)
            result = modeling.run(project, subject, session, scan, time_series, 
                                  config=MODELING_CONF, technique='mock',
                                  base_dir=self.base_dir)
            # Find the modeling resource.
            rsc = xnat.find_one(project, subject, session, scan=scan,
                                resource=result)
            try:
                assert_is_not_none(rsc, "The %s %s Scan %d %s resource was not"
                                        " created" %
                                        (subject, session, scan, result))
            finally:
                xnat.delete(project, subject)


if __name__ == "__main__":
    import nose

    nose.main(defaultTest=__name__)



from qiprofile_rest_client.helpers import database
from qiprofile_rest_client.model.subject import Subject
from qiprofile_rest_client.model.imaging import Session
from . import (clinical, imaging)


def sync_session(project, collection, subject, session, filename):
    """
    Updates the qiprofile database from the XNAT database content for
    the given session.

    :param project: the XNAT project name
    :param collection: the image collection name
    :param subject: the subject number
    :param session: the XNAT session number
    :param filename: the XLS input file location
    """
    # Get or create the subject database subject.
    key = dict(project=project, collection=collection, number=subject)
    sbj = database.get_or_create(Subject, key)
    # Update the clinical information from the XLS input.
    clinical.sync(sbj, filename)
    # Update the imaging information from XNAT.
    imaging.sync(sbj, session)



import os
from nose.tools import (assert_true, assert_is_not_none)
from nipype.interfaces.dcmstack import MergeNifti
import qixnat
from qipipe.pipeline import (mask, qipipeline)
from ... import ROOT
from ...helpers.logging import logger
from ...unit.pipeline.volume_test_base import VolumeTestBase

FIXTURES = os.path.join(ROOT, 'fixtures', 'staged')
"""The test fixtures directory."""

MASK_CONF = os.path.join(ROOT, 'conf', 'mask.cfg')
"""The test mask configuration."""

RESULTS = os.path.join(ROOT, 'results', 'pipeline', 'mask')
"""The test results directory."""


class TestMaskWorkflow(VolumeTestBase):
    """
    Mask workflow unit tests.
    
    This test exercises the mask workflow on three volumes of one visit in each
    of the Breast and Sarcoma studies.
    """

    def __init__(self):
        super(TestMaskWorkflow, self).__init__(logger(__name__), FIXTURES,
                                               RESULTS)

    def test_breast(self):
        for args in self.stage('Breast'):
            self._test_workflow(*args)

    def test_sarcoma(self):
        for args in self.stage('Sarcoma'):
            self._test_workflow(*args)

    def _test_workflow(self, project, subject, session, scan, *images):
        """
        Executes :meth:`qipipe.pipeline.mask.run` on the input scans.
        
        :param xnat: the XNAT facade instance
        :param project: the input project name
        :param subject: the input subject name
        :param session: the input session name
        :param scan: the input scan number
        :param images: the input 3D NiFTI images to model
        """
        # Make the 4D time series from the test fixture inputs.
        merge = MergeNifti(in_files=list(images),
                           out_format=qipipeline.SCAN_TS_RSC)
        time_series = merge.run().outputs.out_file
        logger(__name__).debug("Testing the mask workflow on the %s %s"
                               " Scan %d time series %s..." %
                               (subject, session, scan, time_series))
        # Run the workflow.
        with qixnat.connect() as xnat:
            xnat.delete(project, subject)
            result = mask.run(project, subject, session, scan, time_series, 
                              base_dir=self.base_dir, config=MASK_CONF)
            # Find the mask resource.
            rsc = xnat.find_one(project, subject, session, scan=scan,
                                resource=result)
            try:
                assert_is_not_none(rsc, "The %s %s Scan %d %s resource was not"
                                        " created" %
                                        (subject, session, scan, result))
            finally:
                xnat.delete(project, subject)

if __name__ == "__main__":
    import nose

    nose.main(defaultTest=__name__)



import os
import re
import glob
import shutil
from nose.tools import (assert_equal, assert_is_not_none)
import nipype.pipeline.engine as pe
from qipipe.pipeline import registration
from ... import (ROOT, PROJECT)
from ...helpers.logging import logger
from ...helpers.name_generator import generate_unique_name
from ...unit.pipeline.volume_test_base import VolumeTestBase

REG_CONF = os.path.join(ROOT, 'conf', 'registration.cfg')
"""The test registration workflow configuration."""

FIXTURES = os.path.join(ROOT, 'fixtures', 'registration')
"""The test fixtures directory."""

RESULTS = os.path.join(ROOT, 'results', 'pipeline', 'registration')
"""The test results directory."""

RESOURCE = generate_unique_name(__name__)
"""The XNAT registration resource name."""


class TestRegistrationWorkflow(VolumeTestBase):
    """
    Registration workflow unit tests.

    This test exercises the registration workflow on three volumes of one visit
    in each of the Breast and Sarcoma studies.
    
    The :class:`qiprofile.pipeline.registration.RegistrationWorkflow` technique
    is set to ``mock``, which mocks the registration process by simply copying
    the input scan file to the output "realigned" file. 
    """

    def __init__(self):
        super(TestRegistrationWorkflow, self).__init__(logger(__name__), FIXTURES,
                                                       RESULTS, use_mask=True)

    def test_breast(self):
        for args in self.stage('Breast'):
            self._test_workflow(*args)

    def test_sarcoma(self):
        for args in self.stage('Sarcoma'):
            self._test_workflow(*args)

    def _test_workflow(self, project, subject, session, scan, *images):
        """
        Executes :meth:`qipipe.pipeline.registration.run` on the given input.

        :param project: the input project name
        :param subject: the input subject name
        :param session: the input session name
        :param scan: the input scan number
        :param images: the input 3D NiFTI images to register
        """
        # A reasonable bolus uptake index.
        bolus_arv_ndx = min(3, len(images) / 3)
        # The target location.
        dest = os.path.join(RESULTS, subject, session, 'scans', str(scan),
                            'registration', RESOURCE)
        logger(__name__).debug("Testing the registration workflow on %s %s"
                               " Scan %d..." %
                               (subject, session, scan))
        with qixnat.connect() as xnat:
            xnat.delete(project, subject)
            result = registration.run(project, subject, session, scan,
                                      bolus_arv_ndx, *images,
                                      config=REG_CONF, resource=RESOURCE,
                                      technique='mock', dest=self.dest,
                                      base_dir=self.base_dir)
            # Verify the result.
            try:
                self._verify_result(xnat, subject, session, scan, result)
            finally:
                xnat.delete(project, subject)

    def _verify_result(self, xnat, subject, session, scan, result):
        """
        :param xnat: the XNAT connection
        :param subject: the registration subject
        :param session: the registration session
        :param scan: the input scan number
        :param result: the meth:`qipipe.pipeline.registration.run` result
            output file paths
        """
        # Verify that the XNAT resource object was created.
        rsc = xnat.find_one(PROJECT, subject, session, scan=scan,
                            resource=RESOURCE)
        assert_is_not_none(rsc,  "The %s %s Scan %d %s XNAT registration"
                                 " resource object was not created" %
                                 (subject, session, scan, RESOURCE))

        # Verify that the registration result is accurate.
        split = (os.path.split(location) for location in result)
        out_dirs, out_files = (set(files) for files in zip(*split))
        rsc_files = set(rsc_obj.files().get())
        assert_equal(out_dirs, set([self.dest]),
                     "The %s %s Scan %d %s XNAT registration result directory"
                      " is incorrect - expected %s, found %s" %
                      (subject, session, scan, RESOURCE, self.dest, out_dirs))
        assert_equal(out_files, rsc_files,
                     "The %s %s Scan %d %s XNAT registration result file names"
                     " are incorrect - expected %s, found %s" %
                     (subject, session, scan, RESOURCE, rsc_files, out_files))

        # Verify that the output files were created.
        dest_files = (os.path.join(self.dest, location)
                      for location in os.listdir(self.dest))
        assert_equal(set(dest_files), set(result),
                     "The %s %s Scan %d %s XNAT registration result is"
                     " incorrect: %s" %
                     (subject, session, scan, RESOURCE, result))


if __name__ == "__main__":
    import nose

    nose.main(defaultTest=__name__)



import os
import shutil
from nose.tools import (assert_true, assert_is_not_none)
from qipipe.pipeline import staging
import qixnat
from qipipe.staging.iterator import iter_stage
from ... import (ROOT, PROJECT)
from ...helpers.logging import logger
from ...helpers.staging import subject_sources

FIXTURES = os.path.join(ROOT, 'fixtures', 'staging')
"""The test fixture directory."""

RESULTS = os.path.join(ROOT, 'results', 'pipeline', 'staging')
"""The test results directory."""


class TestStagingWorkflow(object):
    """Staging workflow unit tests."""

    def setUp(self):
        shutil.rmtree(RESULTS, True)

    def tearDown(self):
        shutil.rmtree(RESULTS, True)

    def test_breast(self):
        self._test_collection('Breast')
    
    def test_sarcoma(self):
        self._test_collection('Sarcoma')

    def _test_collection(self, collection):
        """
        Run the staging workflow on the given collection and verify that
        the sessions are created in XNAT.

        :Note: This test does not verify the CTP staging area nor the
        uploaded image file content. These features should be verified
        manually.
        
        :param collection: the image collection name
        """
        fixture = os.path.join(FIXTURES, collection.lower())
        logger(__name__).debug("Testing the staging workflow on %s..." %
                               fixture)

        # The staging destination and work area.
        dest = os.path.join(RESULTS, 'staged')
        work = os.path.join(RESULTS, 'work')

        # The test subject => directory dictionary.
        sbj_dir_dict = subject_sources(collection, fixture)
        # The test subjects.
        subjects = sbj_dir_dict.keys()
        # The test source directories.
        inputs = sbj_dir_dict.values()

        with qixnat.connect() as xnat:
            # Delete any existing test subjects.
            for sbj in subjects:
                xnat.delete(PROJECT, sbj)
            # Run the workflow on each session fixture.
            for scan_input in iter_stage(PROJECT, collection, *inputs, dest=dest):
                work_dir = os.path.join(work, 'scan', str(scan_input.scan))
                stg_wf = staging.StagingWorkflow(project=PROJECT,
                                                 base_dir=work_dir)
                stg_wf.set_inputs(scan_input, dest=dest)
                stg_wf.run()
                # Verify the result.
                sess_obj = xnat.find_one(PROJECT, scan_input.subject,
                                            scan_input.session)
                assert_is_not_none(sess_obj.exists(),
                                   "The %s %s session was not created in XNAT" %
                                   (scan_input.subject, scan_input.session))
                sess_dest = os.path.join(dest, scan_input.subject, scan_input.session)
                assert_true(os.path.exists(sess_dest), "The staging area"
                            " was not created: %s" % sess_dest)
                # The XNAT scan object.
                scan_obj = xnat.find_one(PROJECT, scan_input.subject,
                                         scan_input.session, scan=scan_input.scan)
                assert_is_not_none(scan_obj,
                                   "The %s %s scan %s was not created in XNAT" %
                                   (scan_input.subject, scan_input.session, scan_input.scan))
                # The XNAT NiFTI resource object.
                rsc_obj = scan_obj.resource('NIFTI')
                assert_true(rsc_obj.exists(),
                            "The %s %s scan %s %s resource was not created in XNAT" %
                            (scan_input.subject, scan_input.session,
                             scan_input.scan, 'NIFTI'))
                for volume in scan_input.iterators.dicom.iterkeys():
                    fname = "volume%03d.nii.gz" % volume
                    file_obj = rsc_obj.file(fname)
                    assert_true(file_obj.exists(),
                                "The %s %s scan %s file %s was not created in XNAT" %
                                (scan_input.subject, scan_input.session,
                                 scan_input.scan, fname))

            # Delete the test subjects.
            for sbj in subjects:
                xnat.delete(PROJECT, sbj)


if __name__ == "__main__":
    import nose

    nose.main(defaultTest=__name__)
